#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "pvm3.h"

#define MAX_DEPTH           75
#define MAX_PROCESSORS      20
#define MAXOPS              4
#define MAXOPS2             16
#define THRESHOLD_INCREMENT 2
#define LISTINITSIZE        500
#define JOBTHRESHOLD        1
#define WAITTIME            1
#define SAMPLESIZE          200000

enum {PUZZLE_COMING, GOAL_COMING, THRESHOLD_COMING, NODE_REQUEST,
      LOAD_BAL_SWITCH_COMING, ORDERING_SWITCH_COMING, REJECT, GOAL_INFO,
      OPORDER_COMING, HALT_PROCESSING, NO_MESSAGE, BUSY, IDLE, EUREKA,
      GIVEAWAY_SWITCH_COMING, ALPHA_COMING, GOFORIT, CLUSTER_ROOT,
      CLUSTER_SIZE, SEND_GOAL_INFO, ANTICIPATORY_SWITCH_COMING,
      GOAL_LEVEL_COMING, STATS_COMING, NODE_INFO,
      LOWEST_HEURISTIC_COMING, MOVING_ON} message_type;
enum {KUMAR_RAO, DTS, PWS, CDISTRIB} distribution_types;
enum {TOIDA,  LOCAL, NONE, CORDER} ordering_types;
enum {BOTTOM, TOP, CGIVEAWAY} giveaway_types;
enum {OFF, NEIGHBOR, RANDOM, ASYNC, CLOADBAL} loadbal_types;
enum {OPTIMAL, SUBOPTIMAL} optimal_types;
enum {FALSE, TRUE} boolean_types;

typedef struct {
   int puzzle[16], parent, child_expanded[MAXOPS];
   int level, h, cum_heuristic_error, path[MAX_DEPTH], number_children;
} Node;

typedef struct {
   Node *Elements;          /* Set of elements in the list */
   int First;               /* First element in list */
   int Last;                /* Last element in list */
   int MaxSize;             /* Number of allocated elements in this list */
} List;

/*******************************************************************************

   The message_type variable contains all of the possible messages
   that can be passed from/to a processor node.  These are defined
   as:

   PUZZLE_COMING              - A puzzle is either coming from the host or from
		                another node

   GOAL_COMING                - A goal configuration is coming from the host

   THRESHOLD_COMING           - A new threshold is coming from the host 
				processor

   NODE_REQUEST               - A processor has run out of work and has sent 
		                a request for work to another processor

   LOAD_BAL_SWITCH_COMING     - The host sends a message of this type to 
			        each processor indicating the approach to
			        use for load balancing - see loadbal_types

   ORDERING_SWITCH_COMING     - The host sends a message of this type to
			        each processor indicating the approach to
			        use for ordering - see ordering_types

   REJECT                     - In response to a NODE_REQUEST message, a 
		                processor may have no work to give and thus
		                sends a reject message.

   GOAL_INFO                  - A processor that finds a goal node sends
                                information about the goal node to the host.

   OPORDER_COMING             - Like ORDERING_SWITCH_COMING, this indicates
			        the ordering method to be used.

   HALT_PROCESSING            - A processor has found a goal node and has
		                notified the host.  The host then sends out
		                a halt message to stop all work.

   NO_MESSAGE                 - A dummy message used in the program.

   BUSY                       - A status message sent back to the host 
		                indicating that this processor is busy.

   IDLE                       - A status message like BUSY indicating an idle
		                processor.  When the IDLE processor count
		                reaches the number of processors, then the
		                threshold is updated.

   EUREKA                     - A message sent to the host indicating that a
				goal was found on a processor.

   GIVEAWAY_SWITCH_COMING     - The host sends a message of this type to
			        each processor.  When a NODE_REQUEST message
			        is received work can be taken from the top
			        or bottom of the stack.  This message sets
			        this for each processor.  

   ALPHA_COMING               - Like ORDERING_SWITCH_COMING, this indicates
			        the amount of the work stack to be distributed
			        when a NODE_REQUEST is processed.

   GOFORIT                    - This is a message indicating that tells all 
			        processors to continue after a blocked message

   CLUSTER_ROOT               - This is a message sent by the host to all 
			        processors used in conjunction with clustering
			        indicating the lowest processor number in the
			        cluster.  

   CLUSTER_SIZE               - This is also used in conjunction with clustering
			        indicating the number of processors in each
			        cluster.

   SEND_GOAL_INFO             - This is sent by the host to the processor
			        locating a goal indicating that the goal
			        information is to be sent.
 
   ANTICIPATORY_SWTICH_COMING - This is a message sent by the host controling
				the anticipatory value.  When the size of the
				stack falls below this value, a work request
				is generated.

   GOAL_LEVEL_COMING          - Inform the processors of the least-cost goal
                                found so far.

   STATS_COMING               - Provide a description of this subtree to the
                                host.

   NODE_INFO                  - The host sends information about the array of
                                contributing processors to each processor.

   LOWEST_HEURISTIC_COMING    - This is also sent after a processor goes idle
				indicating the lowest heuristic generated.

   MOVING_ON                  - This is sent by the host telling a processor
				to proceed.

   PVM_EXIT                   - Tell each processor to finish execution.
*******************************************************************************/

void Insert();
int  EmptyList();
float stddev();
List *MakeEmptyList(), *RestoreList();
Node RemoveFirst(), RemoveLast();

char datafile[50], lfile[50], cfile[50];
int message, lowest_h_path[MAX_DEPTH], node_count[MAXOPS], root_heuristic;
int tids[MAX_PROCESSORS], host, puzzle[16], goal[16];
int threshold, initial_threshold, type, source, rc;
int lowest_heuristic, lowest_h_level, lowest_cum_error, busy_message_sent;
int idle_message_sent, waiting_on_request, number_of_rejects, distribution_type;
int load_balancing, ordering, giveaway, anticipatory;
int oporder[MAXOPS], initial_oporder[MAXOPS];
int goal_oporder[MAXOPS], goal_path[MAX_DEPTH], numnodes;
int optimal, goal_level, goal_number_nodes, total_nodes, parents, prev_nodes;
float alpha, bf, hbf, avg_herror;
List *OPEN;
